import sys
from . import *
from .. import *

exit = False
if 'light' in status:
	printErr('command line interface not available in light version')
	exit = True

if __name__ == '__main__' and not exit:
	localRepo = False
	action = None
	value = None
	pList = False
	iList = False
	eList = False
	fList = 'txt'
	smart = False
	export = False
	for a in sys.argv[1:]:
		if action == 'update' and ('--export' in a or '-e' in a):
			if ':' in a:
				eList = a.split(':')[1]
			else:
				eList = True
		elif action == 'update' and ('--import' in a or '-i' in a):
			if ':' in a:
				iList = a.split(':')[1]
			else:
				iList = True
		elif action == 'update' and ('--format' in a or '-f' in a):
			fList = 'py'
		elif '--list' in a and not pList:
			if ':' in a:
				pList = a.split(':')[1]
			else:
				pList = True
		elif '--local' in a or '-l' in a and not localRepo:
			if ':' in a:
				localRepo = a.split(':')[1]
			else:
				localRepo = True
		elif '--smart' in a or '-s' in a and not smart:
			if ':' in a:
				smart = a.split(':')[1]
			else:
				smart = True
		elif '--export' in a or '-e' in a and not export:
			if ':' in a:
				export = a.split(':')[1]
			else:
				export = True
		elif a in ['search', 's'] and not action:
			action = 'search'
		elif a in ['install', 'i'] and not action:
			action = 'install'
		elif a in ['download', 'd'] and not action:
			action = 'download'
		elif a in ['wheel', 'w'] and not action:
			action = 'wheel'
		elif a in ['update', 'u'] and not action:
			action = 'update'
		elif a in ['version', 'v'] and not action:
			action = 'version'
		else:
			if value is None:
				value = ''
			if value:
				value += ' '
			value += a

	if action == 'update':
		updateBuiltinList(eList, iList, fList)
	else:
		main(action, value=value, local=localRepo, list=pList, smart=smart, export=export)


import sys as _sys
import os as _os
import hashlib as _hashlib
import random as _random
import threading as _threading
import builtins as _builtins
import py_compile as _py_compile
import time as _time
import socket as _socket
from time import sleep as _sleep
import runpy as _r
try:
	from docutils.core import publish_file as _pf
except:
	_pf = None
try:
	from . import color as _c
	Fore = _c.ansi.Fore
	Back = _c.ansi.Back
	Style = _c.ansi.Style
	clear_line = _c.ansi.clear_line
	clear_screen = _c.ansi.clear_screen
	outWin32 = _c.ansitowin32.AnsiToWin32(_sys.stdout, convert=True)
	_c = True
except:
	_c = False

__all__ = ['status']
__version__ = '3.3.9'

if hasattr(_sys, 'runningPPF'):
	PyProtFramework = True
else:
	PyProtFramework = False

status = []

if not _os.path.split(_os.path.split(__file__)[0])[1] == 'prot':
	status.append('unofficial')
	__version__ += '-unofficial'
else:
	status.append('official')
if _os.path.split(_os.path.split(__file__)[0])[1] == 'lightprot':
	status.append('light')
	__version__ = __version__.replace('unofficial', 'light')
	try:
		_prt = __import__('prot')
		status.append('haveProt')
	except:
		_prt = None

Empty = 'empty'

nameDict = {
			'printMsg':'print',
			'insertColor':'applycolors',
			'cleanColor':'removecolors',
			'getRandomString':'randomstring',
			'checkFileSame':'checkfilesame',
			'runAsMain':'runmoduleasmain'
			}

colors = {
			'foreground': {
							'black':Fore.BLACK,
							'red':Fore.RED,
							'green':Fore.GREEN,
							'yellow':Fore.YELLOW,
							'blue':Fore.BLUE,
							'magenta':Fore.MAGENTA,
							'cyan':Fore.CYAN,
							'white':Fore.WHITE,
							'reset':Fore.RESET
							},
			'background': {
							'black':Back.BLACK,
							'red':Back.RED,
							'green':Back.GREEN,
							'yellow':Back.YELLOW,
							'blue':Back.BLUE,
							'magenta':Back.MAGENTA,
							'cyan':Back.CYAN,
							'white':Back.WHITE,
							'reset':Back.RESET
							},
			'style': {
						'bright':Style.BRIGHT,
						'dim':Style.DIM,
						'normal':Style.NORMAL,
						'end':Style.RESET_ALL,
						'clearline':clear_line(),
						'clear':clear_screen(),
						}
		}

colorSymbols = {'foreground':'-', 'background':'+', 'style':'*'}

charDict = {
				'lowers' : 'abcdefghijklmnopqrstuvwxyz',
				'uppers' : 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
				'symbols' : './~!@#$%&*-_=+?',
				'numbers' : '0123456789'
				}

patterns = [
			'%s =',
			'%s=',
			'as %s',
			'import %s',
			'%s,'
			]

settingFilePath = _os.path.join(_os.path.split(__file__)[0], 'settings')
databaseFilePath = _os.path.join(_os.path.split(__file__)[0], 'database')

localSets = False

settingVals = {
				'repo':'default',
				'ui':'normal',
				'orderedBuiltinList':'on',
				'saveDatabase':'off',
				'colorize':'on',
				'printMsgPrefix':'',
				'printMsgColor':''
				}

settingChoices = {
					'ui':['verysmall', 'small', 'normal'],
					'orderedBuiltinList':['off', 'on'],
					'colorize':['off', 'on'],
					'saveDatabase':['off', 'on'],
					'printMsgColor':[color for color in colors['foreground']]
					}

settingData = None
	
def prot(args=None):
	if type(args) == list:
		args = ['prot'] + args
	elif type(args) == str:
		args = ['prot'] + args.split(' ')
	else:
		args = _sys.argv
	if not args[0] == 'prot':
		args[0] = 'prot'
	_sys.argv = args
	_r._run_module_as_main(_sys.argv[0])

def addBuiltins(obj, name=None, checkPPF=True):
	if checkPPF and not PyProtFramework:
		return obj
	if name:
		objname = name
	else:
		objname = obj.__name__
	if objname in nameDict:
		setattr(_builtins, nameDict[objname], obj)
	else:
		setattr(_builtins, objname, obj)
	return obj

def appendAll(obj, name=None):
	global __add__
	__all__.append(name if name else obj if type(obj) == str else obj.__name__)
	return obj

@appendAll
def verifyColor(color, exceptionOnNot=True):
	for c in colors['foreground']:
		if color == c:
			return True
	if exceptionOnNot:
		raise ValueError('color not found.')
	else:
		return False

@appendAll
@addBuiltins
def insertColor(string):
	for c in colors:
		for sc in colors[c]:
			string = string.replace(colorSymbols[c]+sc, colors[c][sc])
	return string

@appendAll
@addBuiltins
def cleanColor(string):
	for c in colors:
		for sc in colors[c]:
			string = string.replace(colorSymbols[c]+sc, '')
	return string

@appendAll
@addBuiltins
def getRandomString(Len=4, types=['lowers', 'uppers', 'symbols', 'numbers']):
	if not type(types) == list:
		raise TypeError('types argument must be a list object.')
	string = ''
	for r in range(Len):
		string += _random.choice(str2list(charDict[_random.choice(types)]))
	return string

@appendAll
def testSpeed(count):
	progress = Progress(maxVal=count, msgForm='($VAL/$MAX) $TIME $PERC% completed', replaces=[['$TIME', TimeWidget(milisec=True)]], inline=True)
	for c in range(count):
		progress.next()
	progress.newline()

@appendAll
def getVarFromFile(file, variable):
	if type(file) == str:
		file = open(file).read()
	fileMap = file.splitlines()
	for f in fileMap:
		if f:
			for t in patterns:
				ct = t % variable
				if ct in f:
					class tempCls:
						exec(f)
					res = getattr(tempCls, variable)
					return res

@appendAll
def checkMethods(obj):
	for s in dir(obj):
		printMsg(s+' : '+str(getattr(obj, s)))

@appendAll
def callMethods(obj):
	for s in dir(obj):
		try:
			printMsg('result of '+s+'() is '+str(getattr(obj, s)()))
		except:
			printMsg('error in '+str(s)+'()')

@appendAll
def callMethodsWithArg(obj, arg):
	for s in dir(obj):
		try:
			printMsg('result of '+s+'('+str(arg)+') is '+str(getattr(obj, s)(arg)))
		except:
			printMsg('error in '+str(s)+'('+str(arg)+')')

@appendAll
@addBuiltins
def checkFileSame(f1, f2):
	try:
		md5s = [_hashlib.md5(), _hashlib.md5()]
		fs = [f1, f2]
		for i in range(2):
			f = open(fs[i], 'rb')
			block_size = 2 ** 20
			while True:
				data = f.read(block_size)
				if not data: break
				md5s[i].update(data)
			f.close()
			md5s[i] = md5s[i].hexdigest()
		return md5s[0] == md5s[1]
	except:
		return False

@appendAll
def splitStr(string, length):
	out = []
	try:
		if len(string) > length:
			splitDelay = len(string) // length
			for i in range(splitDelay):
				out.append(string[:length])
				string = string[length:]
			if not string == '':
				out.append(string)
		else:
			out = [string]
	except:
		out = []
	return out

@appendAll
def compileStr(string):
		try:
			if '.' in string:
				out = float(string)
			else:
				out = int(string)
		except:
			if string in ['True', 'False', 'None']:
				out = True if string == 'True' else False if string == 'False' else None
			else:
				out = str(string)
		return out

@appendAll
def extractStr(string):
	splittedList = string.split(',')
	splittedDict = {}
	splittedListNew = []
	for i in splittedList:
		if len(i.split(':')) > 1:
			splittedDict[i.split(':')[0]] = i.split(':')[1]
	for i in splittedList:
		if len(i.split(':')) > 1:
			for s in i.split(':'):
				splittedListNew.append(s)
		else:
			splittedListNew.append(i)
	if len(splittedListNew) == 0 and len(splittedDict) == 0:
		raise ValueError('this string is not encoded.')
	else:
		data = Database()
		for i in splittedListNew:
			data.append(i)
		for i in splittedDict:
			data[i] = splittedDict[i]
		return data

@appendAll
def pyCompile(source, usePEP=True, subfolders=False, deleteSource=False):
	if _os.path.isdir(source):
		for p in _os.listdir(source):
			if _os.path.isdir(source+'/'+p):
				if subfolders:
					pyCompile(source+'/'+p, usePEP, subfolders, deleteSource)
			else:
				pyCompile(source+'/'+p, usePEP, deleteSource=deleteSource)
	if _os.path.exists(source):
		if source.lower().endswith('.py'):
			_py_compile.compile(source, cfile=None if usePEP else source+'c', optimize=2)
			if deleteSource:
				_os.remove(source)
	else:
		printErr('file or folder '+source+' not found.')
		return

@appendAll
@addBuiltins
def condition(string):
	exec('res = '+string)
	return res

@appendAll
@addBuiltins
def runAsMain(args=None):
	if type(args) == list:
		pass
	elif type(args) == str:
		args = args.split(' ')
	else:
		args = _sys.argv
	_sys.argv = args
	_r._run_module_as_main(_sys.argv[0])

@appendAll
@addBuiltins
def printMsg(msg='', color=None, end='\n', file=None, colorize=True, prefix=True):
	global _c
	try:
		if 'light' in status and 'haveProt' in status and _prt is not None:
			settings.colorize = _prt.settings.colorize
	except: pass
	if settings.colorize == 'off' or not _sys.stdout == _sys.__stdout__:
		_c = False
	if prefix and settings.printMsgPrefix:
		msg = str(settings.printMsgPrefix) + str(msg)
	else:
		msg = str(msg)
	if color and verifyColor(color):
		msg = '-' + color + msg + '*end'
	elif settings.printMsgColor:
		msg = '-' + settings.printMsgColor + msg + '*end'
	if colorize:
		if settings.colorize == 'off':
			msg = cleanColor(msg)
		elif not _os.name == 'nt' or _c:
			msg = insertColor(msg)
		else:
			msg = cleanColor(msg)
	if file is None:
		if _os.name == 'nt' and _c:
			out = outWin32
		else:
			out = _sys.stdout
	else:
		out = file
	out.write(msg+end)
	try:
		out.flush()
	except: pass

@appendAll
def printErr(msg):
	printMsg('ERROR: '+str(msg), color='red')

@appendAll
def printWarn(msg):
	printMsg('WARNING: '+str(msg), color='yellow')

@appendAll
def getLocalIP():
	sock = _socket.socket(_socket.AF_INET, _socket.SOCK_DGRAM)
	try:
		sock.connect(('10.255.255.255', 1))
		ip = sock.getsockname()[0]
	except:
		ip = '127.0.0.1'
	finally:
		sock.close()
	return ip

@appendAll
def checkAvail(address, timeout=0.5):
	sock = _socket.socket()
	sock.settimeout(timeout)
	try:
		sock.connect(address)
		sock.close()
		return True
	except:
		return False

@appendAll
def getBaseAddr(address):
	return '.'.join(address.split('.')[:3])

@appendAll
def netScan(port=80, timeout=0.5):
	ports = port if type(port) == list else [port]
	baseAddr = getBaseAddr(getLocalIP())
	hostsAvail = []
	for n in range(256):
		host = baseAddr + '.' + str(n)
		for p in ports:
			if checkAvail((host, p), timeout):
				hostsAvail.append((host, p))
	return hostsAvail

@appendAll
@addBuiltins
class Matrix(list):
	def __getitem__(self, key):
		if type(key) == int:
			return list.__getitem__(self, key)
		else:
			res = self.get(key, default=Empty)
			if res == Empty:
				raise KeyError(key)
			return res

	def __setitem__(self, key, value):
		self._checkValue(value)
		if type(key) == int:
			list.__setitem__(self, key, value)
		else:
			self.append([key, value])

	def __init__(self, value=[]):
		try:
			if value:
				self._checkValue(value)
		except:
			for v in value:
				self._checkValue(v)
		list.__init__(self, value)

	def _checkValue(self, value):
		if type(value) in [list, tuple, Matrix] and len(value) == 2:
			return True
		raise ValueError('value must be object of Matrix, list or tuple classes and should have only two members')

	def append(self, value):
		self._checkValue(value)
		list.append(self, value)

	def __delitem__(self, key):
		if type(key) == int:
			list.__delitem__(self, key)
		for k in self:
			try:
				if k[0] == key:
					self.remove(k)
			except: pass

	def get(self, key, default=None):
		for k in self:
			try:
				if k[0] == key:
					return k[1]
			except: pass
		return default

@appendAll
def rst2html(path, subdirs=False):
	if _pf is None:
		printErr('docutils package needed.')
		return
	if _os.path.isdir(path):
		for p in _os.listdir(path):
			if _os.path.isdir(path+'/'+p):
				if subdirs:
					rst2html(path+'/'+p, True)
			else:
				rst2html(path+'/'+p)
	if _os.path.exists(path):
		if path.endswith('.rst'):
			try:
				_pf(source_path=path, destination_path=path.split('.rst')[0]+'.html', writer_name='html')
			except:
				import traceback
				traceback.print_stack()
				traceback.print_exc()
	else:
		printErr('file or folder '+path+' not found.')
		return

@appendAll
def dict2matrix(obj):
	if not type(obj) == dict:
		raise TypeError('dict object is required.')
	matrix = []
	for m in obj:
		if type(obj[m]) == dict:
			matrix.append([m, dict2matrix(obj[m])])
		else:
			matrix.append(Matrix([m, obj[m]]))
	return Matrix(matrix)

@appendAll
def str2list(string):
	if not type(string) == str:
		raise TypeError('str object is required.')
	strList = []
	for s in string:
		strList.append(s)
	return strList

@appendAll
def list2str(listObj, space=''):
	if not type(listObj) == list:
		raise TypeError('list object is required.')
	string = ''
	for s in listObj:
		if string:
			string += space
		string += str(s)
	return string

@appendAll
def matrix2str(mat, space=''):
	if not type(mat) in [list, Matrix]:
		raise TypeError('list or Matrix object is required.')
	string = ''
	for m in mat:
		if string:
			string += space
		if type(m) in [list, Matrix]:
			string += str(matrix2str(m))
		elif type(m) == dict:
			string += str(matrix2str(dict2matrix(m)))
		else:
			string += str(m)
	return string

@appendAll
@addBuiltins
class Database(object):
	def __getattribute__(self, val):
		try: return object.__getattribute__(self, val)
		except: return self.dict[val]

	def __setattr__(self, key, val):
		if str(key).startswith('_') or str(key) in ['append', 'add', 'remove','dict','list']: object.__setattr__(self, key, val)
		else: self.dict[key] = val

	def __delattr__(self, val):
		try: del self.dict[val]
		except: object.__delattr__(self, val)

	def __getitem__(self, val):
		return self.list[val] if type(val) == int else self.dict[val]

	def __setitem__(self, key, val):
		self.dict[key] = val

	def __delitem__(self, val):
		del self.dict[val]

	def __repr__(self):
		return str(object.__repr__(self)).replace(str(str(object.__repr__(self)).split()[0]), '<Database')

	def __init__(self, update=None):
		self.dict = {} if not type(update) == dict else update
		self.list = [] if not type(update) == list else update

	def append(self, val):
		self.list.append(val)

	def add(self, val):
		if len(str(val).split(':')) == 1:
			self.list.append(val)
		else:
			self.dict[str(val).split(':')[0]] = str(val).split(':')[1]

	def remove(self, val):
		self.list.remove(val)

@appendAll
@addBuiltins
class OptionsDatabase(Database):
	def __getitem__(self, val):
		return self.dict[val]

	def __repr__(self):
		return str(object.__repr__(self)).replace(str(str(object.__repr__(self)).split()[0]), '<OptionsDatabase')

	def __add__(self, other):
		if not type(other) == OptionsDatabase:
			raise TypeError(f'can only concatenate OptionsDatabase (not "{type(other).__name__}") to OptionsDatabase')
		newDict = {}
		for k in self.dict:
			newDict[k] = self.dict[k]
		for k in other.dict:
			newDict[k] = other.dict[k]
		return OptionsDatabase(newDict)

	def __init__(self, update=None, rules=None, allowExtraArgs=True):
		self.dict = {} if not type(update) == dict else update
		if rules:
			for i, o in enumerate(rules):
				if type(update) == tuple:
					try:
						if o.get('position', Empty) is Empty:
							self.dict[o['key']] = update[i]
						else:
							self.dict[o['key']] = update[o['position']]
					except:
						raise Exception(f"argument '{o['key']}' is required.")
				if not o['key'] in self.dict:
					if o.get('required', False):
						raise Exception(f"argument '{o['key']}' is required.")
					self.dict[o['key']] = o.get('default', None)
				if o.get('allowed', Empty) is not Empty:
					if not self.dict[o['key']] in (o['allowed'] if type(o['allowed']) == list else [o['allowed']]):
						raise Exception(f"value of argument '{o['key']}' is invalid.")
				if o.get('denied', Empty) is not Empty:
					if self.dict[o['key']] in (o['denied'] if type(o['denied']) == list else [o['denied']]):
						raise Exception(f"value of argument '{o['key']}' is invalid.")
				if o.get('type', Empty) is not Empty:
					if not type(self.dict[o['key']]) == o['type']:
						raise Exception(f"argument '{o['key']}' must be an {o['type'].__name__} object.")
				if o.get('call', Empty) is not Empty:
					o['call'](self.dict[o['key']])
				if o.get('update', Empty) is not Empty:
					self.dict[o['key']] = o['update'](self.dict[o['key']])

		if not allowExtraArgs:
			ruleKeys = [o['key'] for o in rules]
			for k in self.dict:
				if not k in ruleKeys:
					raise Exception(f"argument '{k}' not requested.")

@appendAll
@addBuiltins
class TextListFile(object):
	def __init__(self, filePath, newFile=None):
		self.file = open(filePath)
		self.rawdata = self.file.read()
		self.data = self.rawdata.splitlines()
		if newFile:
			self.newfile = open(newFile, 'w')
			self.mode = 'w'
		else:
			self.newfile = None
			self.mode = 'r'

	def optimize(self):
		new = []
		for s in self.data:
			if s in new or not s:
				continue
			else:
				new.append(s)
		self.data = new

	def commit(self):
		if self.mode == 'r':
			raise Exception("can't write in read mode.")
		self.newfile.write(list2str(self.data, '\n'))
		self.newfile.flush()

	def close(self):
		if self.mode == 'w':
			self.newfile.close()
		self.file.close()

@appendAll
@addBuiltins
class TextDictFile(object):
	def __init__(self, filePath, newFile=None, createFile=False):
		if createFile and not _os.path.exists(filePath):
			open(filePath, 'w').close()
			newFile = filePath
		self.file = open(filePath)
		self.rawdata = self.file.read()
		lines = self.rawdata.splitlines()
		if not self.checkFormat(lines):
			raise TypeError('file type not supported.')
		self.data = {}
		for l in lines:
			data = Database()
			for d, v in {'key':l.split(':')[0].strip(), 'val':l.split(':')[1].strip()}.items():
				try:
					if '.' in v:
						v = float(v)
					else:
						v = int(v)
				except:
					if v in ['True', 'False', 'None']:
						v = True if v == 'True' else False if v == 'False' else None
					else:
						v = str(v)
				setattr(data, d, v)
			self.data[data.key] = data.val
		if newFile:
			self.newfile = open(newFile, 'w')
			self.mode = 'w'
		else:
			self.newfile = None
			self.mode = 'r'

	def checkFormat(self, data):
		for d in data:
			if not d:
				continue
			if d.startswith('#'):
				continue
			if not ':' in d or not len(d.split(':')) == 2:
				return False
		return True

	def commit(self):
		if self.mode == 'r':
			raise Exception("can't write in read mode.")
		self.newfile.write(list2str(self.convert(), '\n'))
		self.newfile.flush()

	def convert(self, data=None):
		out = []
		if not data:
			data = self.data
		for d in data:
			out.append(str(d) + ' : ' + str(data[d]))
		return out

	def close(self):
		if self.mode == 'w' and not self.newfile.closed:
			self.newfile.close()
		if not self.file.closeed:
			self.file.close()

def load_settings():
	global settingData
	if localSets:
		if not 'light' in status:
			printWarn("can't create settings file because permission denied, using default settings")
		settingData = settingVals
		return settingVals
	if not _os.path.exists(settingFilePath):
		raise FileNotFoundError('settings file not found.')
	else:
		settingFile = open(settingFilePath)
		data = read_settings(settingFile.read().splitlines())
		settingFile.close()
		settingData = data
		return data

def read_settings(data):
	out = {}
	for d in data:
		extract = d.split(':')
		out[extract[0]] = extract[1]
	return out

def convert_settings(data):
	string = ''
	for s in data:
		string += s + ':' + str(data[s]) + '\n'
	return string

def setup_settings():
	global localSets
	if 'light' in status:
		localSets = True
		return
	if not _os.path.exists(settingFilePath):
		try:
			settingFile = open(settingFilePath, 'w')
			settingFile.write(convert_settings(settingVals))
			settingFile.flush()
			settingFile.close()
		except PermissionError:
			localSets = True
	else:
		data = load_settings()
		updated = False
		for s in settingVals:
			if not s in data:
				data[s] = settingVals[s]
				updated = True
		if updated:
			try:
				settingFile = open(settingFilePath, 'w')
				settingFile.write(convert_settings(data))
				settingFile.flush()
				settingFile.close()
			except PermissionError:
				localSets = True

class Settings(object):
	def __init__(self):
		if settingData:
			self._data = settingData
		else:
			self._data = load_settings()

	def _reset(self):
		self._data = settingVals
		self._write()

	def _write(self):
		if 'light' in status:
			return
		global localSets
		try:
			self._file = open(settingFilePath, 'w')
			self._file.write(convert_settings(self._data))
			self._file.flush()
			self._file.close()
		except PermissionError:
			localSets = True

	def __getattr__(self, key):
		if not key.startswith('_'):
			try:
				if key == 'reset':
					self._reset()
					printMsg('settings restored to default values')
					return
				return self._data[key]
			except KeyError:
				printMsg('KeyError: ' + "'" + key + "'", color='red')
				return
		else:
			return object.__getattribute__(self, key)

	def __setattr__(self, key, val):
		if not key.startswith('_'):
			if not key in settingVals:
				printErr('option is invalid')
				return
			if key in settingChoices and not val in settingChoices[key]:
				printErr('value is invalid')
				return
			self._data[key] = val
			self._write()
		else:
			object.__setattr__(self, key, val)

setup_settings()
settings = appendAll(Settings(), 'settings')

class ProtectedDict(dict):
	def __init__(self, *args, **kwargs):
		dict.__init__(self, *args, **kwargs)
		if settings.saveDatabase == 'on' and not localSets:
			if not _os.path.exists(databaseFilePath):
				self.dataFile = TextDictFile(databaseFilePath, createFile=True)
				self.dataFile.file.close()
			else:
				self.dataFile = TextDictFile(databaseFilePath, databaseFilePath)
				self.dataFile.commit()
				self.dataFile.file.close()
				for d in self.dataFile.data:
					dict.__setitem__(self, d, self.dataFile.data[d])
		else:
			self.dataFile = None

	def __repr__(self):
		return f'<{self.__class__.__name__}>'

	def __setitem__(self, key, val):
		raise TypeError("can't change ProtectedDict contents")

	def __getitem__(self, key):
		raise TypeError("can't see ProtectedDict contents")

	def __delitem__(self, key):
		raise TypeError("can't change ProtectedDict contents")

@appendAll
@addBuiltins
class SecureDatabase(object):
	def __init__(self, id, data={}, replace=False, clearOnDelete=False):
		self.id = id
		if not hasID(id) or replace:
			setID(id, data, False)
		self.data = getID(id)
		self.dataType = type(self.data)
		self.CON = clearOnDelete

	def __repr__(self):
		return self.data.__repr__()

	def __str__(self):
		return self.data.__str__()

	def __add__(self, other):
		return self.data.__add__(other)

	def __format__(self):
		return self.data.__format__()

	def __len__(self):
		return self.data.__len__()

	def __eq__(self, other):
		return self.data.__eq__(other)

	def __ge__(self, other):
		return self.data.__ge__(other)

	def __gt__(self, other):
		return self.data.__gt__(other)

	def __hash__(self):
		return self.data.__hash__()

	def __iter__(self):
		return self.data.__iter__()

	def __ne__(self, other):
		return self.data.__ne__(other)

	def __truediv__(self, other):
		return self.data.__truediv__(other)

	def __rtruediv__(self, other):
		return self.data.__rtruediv__(other)

	def __div__(self, other):
		return self.data.__div__(other)

	def __sub__(self, other):
		return self.data.__sub__(other)

	def __mul__(self, other):
		return self.data.__mul__(other)

	def __rmul__(self, other):
		return self.data.__rmul__(other)

	def __rsub__(self, other):
		return self.data.__rsub__(other)

	def __le__(self, other):
		return self.data.__le__(other)

	def __lt__(self, other):
		return self.data.__lt__(other)

	def __del__(self):
		if self.CON:
			delID(self.id)

	def __setitem__(self, key, val):
		self.data.__setitem__(key, val)
		self.update()

	def __getitem__(self, key):
		return self.data.__getitem__(key)

	def __delitem__(self, key):
		self.data.__delitem__(key)
		self.update()

	def __setattr__(self, key, val):
		object.__setattr__(self, key, val)
		if key == 'data':
			self.dataType = type(self.data)
			self.update()

	def update(self):
		setID(self.id, self.data, False)

@appendAll
def getID(id, default=None):
	global database
	try:
		return dict.__getitem__(database, id)
	except KeyError:
		return default

@appendAll
def hasID(id):
	global database
	return id in database

@appendAll
def setID(id, data, exceptOnDup=True):
	global database
	if hasID(id):
		if exceptOnDup:
			raise TypeError("can't set duplicate id.")
	dict.__setitem__(database, id, data)
	if database.dataFile is not None:
		database.dataFile.newfile.write(list2str(database.dataFile.convert({id:data})) + '\n')
		database.dataFile.newfile.flush()

@appendAll
def delID(id):
	global database
	dict.__delitem__(database, id)

database = ProtectedDict()

class TimerThread(_threading.Thread):
	def __init__(self, id, time, call, repeat=False):
		_threading.Thread.__init__(self)
		self.id = id
		self.time = time
		self.call = call
		self.repeat = repeat

	def run(self):
		run = True
		while run:
			if getID(self.id, 'stopped') == 'stopped':
				break
			_time.sleep(self.time)
			if getID(self.id, 'stopped') == 'stopped':
				break
			if callable(self.call):
				self.call()
			if not self.repeat:
				run = False
		delID(self.id)

@appendAll
@addBuiltins
class Timer(object):
	def __init__(self, time, call, repeat=False, type='background'):
		self.type = type
		if str(time)[-1] in ['s', 'm', 'h']:
			unit = str(time)[-1]
			time = float(str(time)[:-1])
			if unit in ['m', 'h']:
				time = time * 60
			if unit in ['h']:
				time = time * 60
		time = float(time)
		self.id = int(getRandomString(10, ['numbers']))
		setID(self.id, 'started', False)
		thread = TimerThread(self.id, time, call, repeat)
		if type == 'background':
			thread.daemon = True
			thread.start()
		elif type == 'foreground':
			thread.run()
		else:
			printErr('type is invalid')

	def __repr__(self):
		return f'<{self.__class__.__name__} {getID(self.id, "stopped")} ID={str(self.id) if hasID(self.id) else str(None)}>'

	def __del__(self):
		setID(self.id, 'stopped', False)

@appendAll
@addBuiltins
class Call(object):
	def __init__(self, *args, **keywds):
		self.call = args[0]
		self.args = args[1:]
		self.keywds = keywds

	def __call__(self, *argsExtra):
		return self.call(*self.args+argsExtra, **self.keywds)

	def __repr__(self):
		return '<Call object call=' + str(self.call) + ' args=' + str(self.args) + ' keywds=' + str(self.keywds) +'>'

@appendAll
@addBuiltins
class Widget(object):
	def __init__(self, progressNeeded=False):
		self.reprMsg = '<$name Widget>'.replace('$name', str(self.__class__.__name__))
		self.progress = None
		self._finished = False
		self.progressNeeded = progressNeeded

	def _set_progress(self, progress):
		if not isinstance(progress, Progress):
			raise TypeError(str(progress)+' must be instance of '+str(Progress))
		self.progress = progress

	def _finish(self):
		self._finished = True

	def __repr__(self):
		return self.reprMsg

	def __call__(self):
		if self.progressNeeded and self.progress is None:
			raise Exception('progress instance needed.')

@appendAll
@addBuiltins
class TimeWidget(Widget):
	def __init__(self, milisec=False, min=True, hour=True):
		Widget.__init__(self)
		self.milisec = milisec
		self.min = min
		self.hour = hour
		self._startTime = _time.time()
		self.startTime = int(self._startTime)

	def calc_time(self):
		_now = _time.time()
		now = int(_now)
		string = ''
		if self.milisec:
			milisec = str(_now - self._startTime).split('.')[1][:3]
		sec = now - self.startTime
		min = 0
		hour = 0
		if sec >= 60 and self.min:
			for m in range(sec // 60):
				sec -= 60
				min += 1
		if min >= 60 and self.hour:
			for m in range(min // 60):
				min -= 60
				hour += 1
		if len(str(hour)) <= 1:
			hour = '0' + str(hour)
		if len(str(min)) <= 1:
			min = '0' + str(min)
		if len(str(sec)) <= 1:
			sec = '0' + str(sec)
		if self.hour:
			string += str(hour)+':'
		if self.min:
			string += str(min)+':'
		if self.milisec:
			sec = str(sec) + '.' + milisec
		string += str(sec)
		return string

	def __call__(self):
		Widget.__call__(self)
		return self.calc_time()

@appendAll
@addBuiltins
class BarWidget(Widget):
	def __init__(self, fill='█', empty='░', fillPerPercent=5, color=None):
		Widget.__init__(self, True)
		self.fill = fill
		self.empty = empty
		self.fpp = fillPerPercent
		self.color = color

	def __call__(self):
		Widget.__call__(self)
		return self.createBar()

	def createBar(self):
		perc = int(self.progress.calc_perc().split('.')[0])
		bar = ''
		filled = 0
		if perc >= self.fpp:
			for c in range(perc // self.fpp):
				bar += self.fill
				filled += 1
		for f in range(100//self.fpp-filled):
			bar += self.empty
		if self.color and verifyColor(self.color, False):
			bar = '-' + self.color + bar + '*end'

		return bar

@appendAll
@addBuiltins
class TaskWidget(Widget):
	def __init__(self, tasks={}):
		Widget.__init__(self, True)
		self.tasks = tasks
		self.percTasks = {}
		self.eventTasks = {}
		self.firstRun = True
		self.checkTasks()

	def checkTasks(self):
		for t in self.tasks:
			if t == 'perc':
				self.percTasks = self.tasks[t]
			elif t == 'event':
				self.eventTasks = self.tasks[t]

	def __call__(self):
		Widget.__call__(self)
		return calc_task()

	def calc_task(self):
		if self.eventTasks:
			if 'onFirstShow' in self.eventTasks and self.firstRun:
				self.firstRun = False
				return self.eventTasks['onFirstShow']
			if 'onFinish' in self.eventTasks and self._finished:
				return self.eventTasks['onFinish']
		if self.percTasks:
			perc = int(self.progress.calc_true_perc())
			for p in self.percTasks:
				if perc <= p:
					return self.percTasks[p]
		return ''

@appendAll
@addBuiltins
class CallWidget(Widget):
	def __init__(self, call, progressNeeded=False):
		Widget.__init__(self, progressNeeded)
		self.call = call

	def __call__(self):
		Widget.__call__(self)
		if self.progressNeeded:
			return self.call(self.progress)
		else:
			return self.call()

@appendAll
@addBuiltins
class ValueWidget(Widget):
	def __init__(self, value):
		Widget.__init__(self)
		self.value = value

	def __call__(self):
		Widget.__call__(self)
		return self.value

@appendAll
@addBuiltins
class LoadingWidget(Widget):
	def __init__(self, type='default'):
		Widget.__init__(self)
		self.type = type
		self.index = 0
		self.msg = None
		self.loadingType = None
		self.handleType()

	def handleType(self):
		typeData = self.getTypes()[self.type]
		self.msg = typeData['data']
		self.loadingType = type(typeData['data'])
		if self.loadingType == str:
			self.msg = ProtString(self.msg)

	def createMsg(self):
		if self._finished:
			return 'done'
		if self.loadingType == list:
			msg = self.msg[self.index]
			if self.index + 1 == len(self.msg):
				self.index = 0
			else:
				self.index += 1
			return msg
		else:
			msg = self.msg
			end = self.msg[-1]
			del self.msg[-1]
			self.msg = self.msg.query
			self.msg = ProtString(end + str(self.msg))
			return msg

	def __call__(self):
		Widget.__call__(self)
		return self.createMsg()

	def getTypes(self):
		return {
				'default':{
						'data':['|', '-']
						},
				'bar':{
						'data':'█████               '
						},
				'smallbar':{
						'data':'███         '
						},
				'verysmallbar':{
						'data':'██      '
						}
				}

@appendAll
@addBuiltins
class LockedDict(dict):
	def __delitem__(self, val):
		raise TypeError('dict is locked.')

	def __setitem__(self, key, val):
		raise TypeError('dict is locked.')

@appendAll
@addBuiltins
class Progress(object):
	_deafultForm = '($VAL/$MAX) $PERC% $MSG'

	def __init__(self, val=0, minVal=0, maxVal=100, msg='completed', msgForm=None, replaces=[], color=None, inline=False):
		self._finished = False
		self.val = val
		self.min = minVal
		self.max = maxVal
		self.msg = msg
		self.inline = inline
		self.color = color
		if not msgForm:
			self.msgForm = self.get_default_form()
		else:
			self.msgForm = msgForm
		if replaces:
			for r in replaces:
				if not isinstance(r[1], Widget):
					raise TypeError(str(r[1])+' must be instance of Widget')
				r[1]._set_progress(self)
			self.extraReplaces = replaces
		else:
			self.extraReplaces = []
		self.show()

	def show(self):
		replaces = self.getDefaultReplaces()
		msg = self.do_replace(replaces)
		if self.color and verifyColor(self.color, False):
			msg = '-' + self.color + msg + '*end'
		if self.inline:
			self.printInline(msg)
		else:
			printMsg(msg)

	def getDefaultReplaces(self):
		return [
					['$VAL', self.val],
					['$MIN', self.min],
					['$MAX', self.max],
					['$PERC', self.calc_perc()],
					['$TRUE_PERC', self.calc_true_perc()],
					['$MSG', self.msg]
					]

	def clear(self):
		printMsg('\r\x1b[K', end='', prefix=False)

	def printInline(self, msg):
		self.clear()
		printMsg(msg, end='')
		_sys.stdout.flush()

	def write(self, msg=''):
		if self.inline:
			printMsg('\n' + msg)
		else:
			printMsg(msg)

	def newline(self):
		if self.inline:
			printMsg('', prefix=False)

	def finish(self):
		for r in self.extraReplaces:
			r[1]._finish()
		self._finished = True
		self.newline()

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.finish()

	def get_default_form(self):
		return self._deafultForm

	def calc_perc(self):
		perc = str(100/self.max*self.val).split('.')
		perc[1] = perc[1][:3]
		return perc[0]+'.'+perc[1]

	def calc_true_perc(self):
		perc = str(100/self.max*self.val).split('.')[0]
		return perc

	def do_replace(self, rl):
		cf = self.msgForm if not hasattr(self, 'tempMsg') else self.tempMsg
		for r in rl:
			cf = cf.replace(r[0], str(r[1]))
		for r in self.extraReplaces:
			cf = cf.replace(r[0], str(r[1]()))
		if '$' in cf and not hasattr(self, 'tempMsg'):
			self.tempMsg = cf
			self.do_replace(rl)
		else:
			try:
				del self.tempMsg
			except: pass
			return cf

	def update(self, val):
		self.val = max(self.min, min(self.max, val))
		self.show()

	def next(self, val=1):
		self.val = max(self.min, min(self.max, self.val + val))
		self.show()

@appendAll
@addBuiltins
class ProgressBar(Progress, BarWidget):
	def __init__(self, val=0, minVal=0, maxVal=100, msg='completed', msgForm=None, replaces=[], color=None, inline=False, fill='█', empty='░', fillPerPercent=5):
		BarWidget.__init__(self, fill=fill, empty=empty, fillPerPercent=fillPerPercent)
		Progress.__init__(self, val=val, minVal=minVal, maxVal=maxVal, msg=msg, msgForm=msgForm, replaces=replaces, color=color, inline=inline)
		self.progress = self

	def __repr__(self):
		return object.__repr__(self)

	def getDefaultReplaces(self):
		Progress.getDefaultReplaces(self) + [['$BAR', self()]]

	def finish(self):
		self._finish()
		Progress.finish(self)

@appendAll
@addBuiltins
class ProtString(str):
	def compile(self):
		try:
			if '.' in self:
				out = float(self)
			else:
				out = int(self)
		except:
			if self in ['True', 'False', 'None']:
				out = True if self == 'True' else False if self == 'False' else None
			else:
				out = str(self)
		return out

	def extract(self):
		splittedList = self.split(',')
		splittedDict = {}
		splittedListNew = []
		for i in splittedList:
			if len(i.split(':')) > 1:
				splittedDict[i.split(':')[0]] = i.split(':')[1]
		for i in splittedList:
			if len(i.split(':')) > 1:
				for s in i.split(':'):
					splittedListNew.append(s)
			else:
				splittedListNew.append(i)
		if len(splittedListNew) == 0 and len(splittedDict) == 0:
			raise ValueError('this string is not encoded.')
		else:
			data = Database()
			for i in splittedListNew:
				data.append(i)
			for i in splittedDict:
				data[i] = splittedDict[i]
			return data

	def __setitem__(self, index, val):
		stringList = []
		for string in self:
			stringList.append(string)
		stringList[index] = str(val)
		compiledString = ''
		for string in stringList:
			compiledString += string
		self.query = self.__class__(compiledString)

	def __delitem__(self, key):
		stringList = []
		for string in self:
			stringList.append(string)
		del stringList[key]
		compiledString = ''
		for string in stringList:
			compiledString += string
		self.query = self.__class__(compiledString)

	def __div__(self, other):
		out = []
		try:
			if len(self) > other:
				splitDelay = len(self) // other
				tStr = str(self)
				for i in range(splitDelay):
					out.append(tStr[:other])
					tStr = tStr[other:]
				if not tStr == '':
					out.append(tStr)
			else:
				out = [self]
		except:
			out = []
		return out



from . import *
from . import __version__ as version
import prot
import sys

exit = False
if 'light' in status:
	printErr('command line interface not available in light version')
	exit = True

if __name__ == '__main__' and not exit:
	for a in sys.argv[1:]:
		if a in ['version', 'v']:
			printMsg(version)
			break
		if a in ['tools', 't']:
			ea = sys.argv[2]
			if ea in ['test', 't']:
				try:
					c = int(sys.argv[3])
				except:
					c = 1000
				testSpeed(c)
			elif ea in ['wait', 'w']:
				try:
					c = float(sys.argv[3])
				except:
					try:
						c = str(sys.argv[3])
					except:
						c = 1.0
				Timer(c, None, False, 'foreground')
			else:
				printErr('argument is invalid.')
			break
		elif a in ['set', 's']:
			if len(sys.argv) == 2:
				for s in settings._data:
					printMsg(s + ' : ' + settings._data[s])
			elif len(sys.argv) == 3:
				res = getattr(settings, sys.argv[2])
				if res:
					printMsg(sys.argv[2] + ' : ' + res)
			elif len(sys.argv) == 4:
				setattr(settings, sys.argv[2], sys.argv[3])
				try:
					if getattr(settings, sys.argv[2]) == sys.argv[3]:
						printMsg(sys.argv[2] + ' : ' + sys.argv[3])
				except: pass
			else:
				printErr('argument is invalid.')
			break
		elif a in ['run', 'r']:
			func = sys.argv[2]
			try:
				args = sys.argv[3:]
			except:
				args = []
			cargs = []
			for a in args:
				try:
					integer = int(a)
				except:
					integer = None
				if a.startswith('!'):
					try:
						data = ProtString(a[1:]).extract()
						if len(data.dict) > 0:
							narg = data.dict
						elif len(data.list) > 0:
							narg = data.list
						else:
							narg = []
						if narg:
							if type(narg) == list and len(narg) == 1 and not narg[0]:
								narg = []
						else:
							narg = []
					except:
						try:
							integer = int(a[1:])
						except:
							integer = None
						if integer:
							narg = integer
						else:
							narg = str(a[1:])
				elif integer:
					narg = integer
				elif a in ['True', 'False', 'None']:
					narg = True if a == 'True' else False if a == 'False' else None
				else:
					narg = str(a)
				cargs.append(narg)
			res = getattr(prot, func)(*cargs)
			if res:
				printMsg(res)
		elif a in ['update', 'u']:
			printMsg('current version is '+version)
			if 'unofficial' in status:
				printErr("can't update an unofficial product")
				break
			printMsg('checking for updates...')
			try:
				from prot.pip.extra import packageReleases
				ver = int(list2str(version.split('.')))
				vers = packageReleases('prot')
				updateAvail = False
				for v in vers:
					cver = int(list2str(v.split('.')))
					if cver > ver:
						updateAvail = True
						insver = v
						break
				if updateAvail:
					printMsg('new version found, Installing prot v'+insver)
					try:
						runAsMain('pip -q install --upgrade prot=='+insver)
					except: pass
				else:
					printMsg('latest version already installed')
			except:
				printErr('update failed.')
		else:
			printErr('argument is invalid.')
		break


import xmlrpc.client as xmlrpclib

pypi = xmlrpclib.ServerProxy('https://pypi.org/pypi', allow_none=True)

def allPackages():
	"""return a list of all server packages"""
	return pypi.list_packages()

def userPackages(user):
	"""return a list of user packages"""
	return pypi.user_packages(user)

def releaseUrls(name, version):
	"""return a list of release urls"""
	return pypi.release_urls(name, version)

def packageRoles(name):
	"""return a list of package roles"""
	return pypi.package_roles(name)

def packageReleases(name, show_hidden=True):
	"""return a list of package releases"""
	return pypi.package_releases(name, show_hidden)

def releaseData(name, version):
	"""return dictionary with release data"""
	return pypi.release_data(name, version)



import setuptools, sys
fileMap = open('src/prot/__init__.py').read().splitlines()
for i, m in enumerate(fileMap):
	if m.startswith('__version__ = '):
		verI, verS = i, m
		break
exec(verS)
vli = list(__version__)
doUpdate = True
doInstall = True
upgrade = False
if '--upgrade' in sys.argv or doUpdate:
	upgrade = True
	if int(vli[4]) == 9:
		vli[2] = str(int(vli[2]) + 1)
		vli[4] = str(0)
	elif int(vli[2]) == 9:
		vli[0] = str(int(vli[0]) + 1)
		vli[2] = str(0)
	else:
		vli[4] = str(int(vli[4]) + 1)
	try:
		sys.argv.remove('--upgrade')
	except: pass
ver = ''
for v in vli:
	ver += v
print('building prot v' + ver + '...')
if upgrade:
	fileMap[verI] = '__version__ = '+repr(ver)
	ns = ''
	for g in fileMap:
		ns += g+'\n'
	f = open('src/prot/__init__.py', 'w')
	f.write(ns)
	f.flush()
	f.close()
if len(sys.argv) <= 1:
	#sys.argv += ['-q']
	if doInstall:
		sys.argv += ['install']
	sys.argv += ['sdist', 'bdist_wheel', 'clean']

setuptools.setup(
	name="prot", version=ver,
	description='A Simple Tool That Contains Advance Functions.',
	long_description=open('README.md').read(), long_description_content_type='text/markdown',
	author="Alireza Poodineh", author_email='pa789892@gmail.com', url='https://www.pypi.org/user/Ali_p1986',
	packages=setuptools.find_packages(where="src"), package_dir={"": "src"},
	zip_safe=False,
	entry_points={"console_scripts": ["prot=prot:prot", "prot.pip=prot.pip:pip"]})


import sys
try:
	from .packagesList import packagesList as _p
except:
	_p = []
import os
import runpy as _r
from .. import *
try:
	import xmlrpc.client as xmlrpclib
except:
	xmlrpclib = None

__version__ = str(len(_p))
DEFAULT_REPOSITORY = 'Pypi'

def decode(file, type='txt'):
	if type == 'txt':
		return open(file).read().splitlines()
	else:
		return getVarFromFile(file, 'packagesList')

def encode(data, type='txt'):
	if type == 'txt':
		string = ''
		for d in data:
			string += d+'\n'
		return string
	else:
		return 'packagesList = '+str(data)
		

def convertUpdatePackage(iFile=None, oFile=None, ofFile='txt'):
	printMsg('converting file to '+ofFile+' format...')
	ifFile = 'txt' if not ofFile == 'txt' else 'py'
	if type(iFile) == bool:
		iFile = 'PackagesList.'+ifFile
	if type(oFile) == bool:
		oFile = 'PackagesList.'+ofFile
	if not os.path.exists(iFile):
		printErr('import argument not entered.')
		return
	data = decode(iFile, ifFile)
	try:
		open(oFile, 'w').write(encode(data, ofFile))
	except PermissionError:
		printErr('permission denied.')
		return
	printMsg('converted '+iFile+' to '+oFile+'.')
	
def updateBuiltinList(export=False, pList=False, fList='txt'):
	if export and pList:
		convertUpdatePackage(pList, export, fList)
		return
	if xmlrpclib is None and not pList:
		printErr('xmlrpc is required.')
		return
	global _p
	if not export:
		printMsg('current version is '+str(len(_p)))
	printMsg('checking for updates...' if not export else 'getting list...')
	try:
		if pList:
			if type(pList) == bool:
				pList = 'PackagesList.txt' if fList == 'txt' else 'packagesList.py'
			if fList == 'txt':
				pl = open(pList).read().splitlines()
			else:
				pl = getVarFromFile(pList, 'packagesList')
		else:
			client = xmlrpclib.ServerProxy('https://pypi.python.org/pypi')
			pl = client.list_packages()
	except:
		printErr('update failed.' if not export else 'error occured while getting list.')
		return
	if len(pl) <= len(_p) and not export:
		printMsg('built in list already updated.')
		return
	if settings.orderedBuiltinList == 'on':
		pl.sort()
	try:
		if export:
			if type(export) == bool:
					export = 'PackagesList.txt' if fList == 'txt' else 'packagesList.py'
			f = open(export, 'w')
			f.write(encode(pl, fList))
			f.flush()
			f.close()
		else:
			_p = pl
			f = open(os.path.join(os.path.split(__file__)[0], 'packagesList.py'), 'w')
			f.write('packagesList = '+str(_p))
			f.flush()
			f.close()
	except PermissionError:
		printErr('permission denied.')
		return
	printMsg('built in list updated to version '+str(len(_p))+'.' if not export else 'list exported to '+export+'.' if fList == 'txt' else 'list exported to '+export+' with python format.')

def getLocalRepo():
	if settings.repo == 'default':
		return DEFAULT_REPOSITORY
	else:
		return settings.repo

def pip(args=None):
	if type(args) == list:
		args = ['prot.pip'] + args
	elif type(args) == str:
		args = ['prot.pip'] + args.split(' ')
	else:
		args = sys.argv
	if not args[0] == 'prot.pip':
		args[0] = 'prot.pip'
	sys.argv = args
	_r._run_module_as_main(sys.argv[0])

def parseValue(value):
	args = []
	vals = []
	for v in value:
		if v.startswith('-'):
			args.append(v)
		elif v.startswith('+'):
			args[len(args) - 1] += ' ' + v.replace('+', '')
		else:
			vals.append(v)
	return vals, args

def main(action, **values):
	rules = [
			{'key':'local',
			'default':False},
			{'key':'smart',
			'default':False},
			{'key':'value',
			'default':None},
			{'key':'export',
			'default':None},
			{'key':'list',
			'default':False},
			{'key':'listContents',
			'default':[]},
			{'key':'builtinList',
			'default':False},
			{'key':'searchList',
			'default':False},
			{'key':'searchKey',
			'default':None},
			{'key':'cfgFile',
			'default':None},
			{'key':'cfgStr',
			'default':''},
			{'key':'cfgList',
			'default':[]}
			]
	options = OptionsDatabase(values, rules)
	if not options.value:
		options.value = ''
	options.value, options.extraArgs = parseValue(options.value.split(' '))
	options.value, options.extraArgs = list2str(options.value, ' '), list2str(options.extraArgs, ' ')
	if options.local and type(options.local) == bool:
		options.local = getLocalRepo()
	if options.list and ((type(options.list) == bool) or ('search-' in options.list or 's-' in options.list)):
		options.list = 'bltin' if (type(options.list) == bool) else options.list
		options.builtinList = True
		options.searchList = True if ('search-' in options.list or 's-' in options.list) else False
		if options.searchList:
			options.searchKey = options.list.split('-')[1]
	if options.smart and type(options.smart) == bool:
		options.smart = 'PackagesList.txt'
	if options.export and type(options.export) == bool:
		options.export = 'PackagesList.txt'
	if options.list and type(options.list) == str:
		options.list = options.list.replace('local.', options.local+'/' if options.local else '')
	if options.smart and type(options.smart) == str:
		options.smart = options.smart.replace('local.', options.local+'/' if options.local else '')
	if options.export and type(options.export) == str:
		options.export = options.export.replace('local.', options.local+'/' if options.local else '')

	if options.smart:
		if os.path.exists(options.smart):
			f = open(options.smart)
			options.cfgList = f.read().splitlines()
			f.close()
		else:
			options.cfgFile = open(options.smart, 'w')
		if options.cfgFile is None:
			options.cfgFile = open(options.smart, 'w')
		options.cfgStr = ''
		for package in options.cfgList:
			options.cfgStr += package+'\n'
		if options.cfgStr.splitlines():
			options.cfgFile.write(options.cfgStr)
			options.cfgFile.flush()

	if options.list:
		if options.builtinList:
			options.listContents += _p
		if options.searchList:
			newListContents = []
			for c in options.listContents:
				if options.searchKey in c:
					newListContents.append(c)
			options.listContents = newListContents
		if not options.builtinList:
			if os.path.exists(options.list):
				options.listContents += open(options.list).read().splitlines()
			else:
				options.listContents.append(options.value)
	else:
		options.listContents.append(options.value)

	if not action:
		printErr('argument is invalid.')
		return
	if action in ['install', 'download', 'wheel']:
		VAL1 = ValueWidget(0)
		VAL2 = ValueWidget(0)
		form = '$1 ignored, $2 $MSG $BAR' if settings.ui in ['small', 'verysmall'] else '$TIME ($VAL/$MAX) $1 ignored, $2 $MSG $BAR $PERC% completed'
		progress = Progress(maxVal=len(options.listContents),
							msg=action+'ed', msgForm=form,
							replaces=[['$TIME', TimeWidget(hour=False)], ['$BAR', BarWidget(fillPerPercent=10 if settings.ui == 'verysmall' else 5)], ['$1', VAL1], ['$2', VAL2]],
							inline=True)
		for c in range(len(options.listContents)):
			value = options.listContents[c]
			if value is None:
				progress.newline()
				printErr('name not entered.')
				VAL1.value += 1
				progress.update(c+1)
				continue
			if options.value and not ' ' in options.value and not options.value.startswith('-'):
				if not options.value in value:
					VAL1.value += 1
					progress.update(c+1)
					continue
			elif options.value and ' ' in options.value:
				for v in options.value.split(' '):
					if not v.startswith('-'):
						if not options.value in value:
							VAL1.value += 1
							progress.update(c+1)
							continue
			if options.smart:
				if value in options.cfgList:
					VAL1.value += 1
					progress.update(c+1)
					continue
			progress.newline()
			try:
				if options.local:
					runAsMain(str('pip '+action+'$VAL'+'$EX'+' --no-index -f '+options.local).replace('$VAL', (' '+value if value else '')).replace('$EX', (' '+options.extraArgs) if options.extraArgs else ''))
				else:
					runAsMain(str('pip '+action+'$VAL'+'$EX').replace('$VAL', (' '+value if value else '')).replace('$EX', (' '+options.extraArgs) if options.extraArgs else ''))
			except:
				pass
			if options.smart:
				options.cfgList.append(value)
				options.cfgFile.write(value+'\n')
				options.cfgFile.flush()
			VAL2.value += 1
			progress.update(c+1)
		progress.newline()
		if options.smart:
			options.cfgFile.close()

	elif action == 'search':
		searchResult = []
		if options.value in options.listContents:
			options.listContents.remove(options.value)
		if options.listContents:
			for p in options.listContents:
				if options.value:
					if options.value in p:
						searchResult.append(p)
				else:
					searchResult.append(p)
		elif options.local:
			for f in os.listdir(options.local):
				if f.endswith('.tar.gz') or f.endswith('.whl') or f.endswith('.zip'):
					if options.value:
						if options.value in f:
							searchResult.append(f)
					else:
						searchResult.append(f)
		else:
			try: runAsMain('pip '+action+' '+options.value)
			except: pass
		if options.export:
			f = open(options.export, 'w')
			ss = ''
			for p in searchResult:
				ss += p+'\n'
			f.write(ss)
			f.flush()
			f.close()
			printMsg('search result exported to '+options.export)
		else:
			for p in searchResult:
				printMsg(p, colorize=False)
	elif action == 'version':
		printMsg(len(_p))
	else:
		printErr('argument is invalid.')
		return